"""Equinox implementation of the Llama model with LoRA support."""

import jax
import jax.numpy as jnp
import equinox as eqx
from typing import Optional, Any, List
import ml_dtypes
import jax.nn.initializers as init


DTYPE_MAP = {
    "float32": jnp.float32,
    "float16": jnp.float16,
    "bfloat16": jnp.bfloat16,
    "int32": jnp.int32,
    "int64": jnp.int64,
}

# Define rematerialization policies
remat_policy = {
    "nothing": jax.checkpoint_policies.nothing_saveable,
    "dots": jax.checkpoint_policies.checkpoint_dots,
    "dots_with_no_batch_dims": jax.checkpoint_policies.checkpoint_dots_with_no_batch_dims,
    "everything": jax.checkpoint_policies.everything_saveable,
}


class LlamaConfig:
    def __init__(self, **kwargs):
        self.vocab_size = kwargs.get("vocab_size", 32000)
        self.hidden_size = kwargs.get("hidden_size", 4096)
        self.intermediate_size = kwargs.get("intermediate_size", 11008)
        self.num_hidden_layers = kwargs.get("num_hidden_layers", 32)
        self.num_attention_heads = kwargs.get("num_attention_heads", 32)
        self.num_key_value_heads = kwargs.get("num_key_value_heads", 32)
        self.max_position_embeddings = kwargs.get(
            "max_position_embeddings", 2048
        )
        self.rms_norm_eps = kwargs.get("rms_norm_eps", 1e-6)

        # New attributes
        self.rope_theta = kwargs.get("rope_theta", 10000.0)
        self.attention_bias = kwargs.get("attention_bias", False)
        self.hidden_act = kwargs.get("hidden_act", "silu")
        self.initializer_range = kwargs.get("initializer_range", 0.02)
        self.use_cache = kwargs.get("use_cache", True)
        self.tie_word_embeddings = kwargs.get("tie_word_embeddings", False)
        self.rope_scaling = kwargs.get("rope_scaling", None)

        # Derived attributes
        self.head_dim = self.hidden_size // self.num_attention_heads
        self.pretraining_tp = kwargs.get("pretraining_tp", 1)

        # Optional attributes
        self.bias = kwargs.get(
            "bias", False
        )  # For compatibility with some attention implementations
        self.rope_type = kwargs.get("rope_type", "default")
        self.partial_rotary_factor = kwargs.get("partial_rotary_factor", 1.0)

        # Dropout rates (usually 0.0 for inference)
        self.attention_dropout = kwargs.get("attention_dropout", 0.0)
        self.hidden_dropout = kwargs.get("hidden_dropout", 0.0)

        # Additional optional parameters
        self.bos_token_id = kwargs.get("bos_token_id", None)
        self.eos_token_id = kwargs.get("eos_token_id", None)
        self.pad_token_id = kwargs.get("pad_token_id", None)
        self.torch_dtype = kwargs.get("torch_dtype", None)

        self.lora_rank = kwargs.get("lora_rank", 0)  # Default 0 means no LoRA

        # Store dtype names as strings
        self._param_dtype = kwargs.get("param_dtype", "bfloat16")
        self._compute_dtype = kwargs.get("compute_dtype", "bfloat16")

    @property
    def param_dtype(self):
        """Gets the parameter dtype, converting from string to JAX dtype."""
        return DTYPE_MAP.get(self._param_dtype, jnp.bfloat16)

    @property
    def compute_dtype(self):
        """Gets the compute dtype, converting from string to JAX dtype."""
        return DTYPE_MAP.get(self._compute_dtype, jnp.bfloat16)

    def to_dict(self):
        """Serializes the configuration to a dictionary."""
        return self.__dict__

    def __repr__(self):
        return f"LlamaConfig({', '.join(f'{k}={v}' for k, v in self.__dict__.items())})"


class LlamaEmbedding(eqx.Module):
    weight: jnp.ndarray
    param_dtype: Any
    compute_dtype: Any

    def __init__(
        self,
        num_embeddings,
        embedding_dim,
        param_dtype=jnp.float32,
        compute_dtype=jnp.float32,
        key=None,
    ):
        self.param_dtype = param_dtype
        self.compute_dtype = compute_dtype
        if key is None:
            key = jax.random.PRNGKey(99)
        self.weight = jax.random.normal(
            key, (num_embeddings, embedding_dim), dtype=self.param_dtype
        )

    def __call__(self, x):
        weight = self.weight.astype(self.compute_dtype)
        embeddings = jnp.take(weight, x, axis=0)
        return embeddings.astype(self.compute_dtype)


# TODO(lora): Need to change this to equnix.linear or define my own quax.
# TODO(lora): Remove static=True to do full fine-tuning.
class LlamaLinear(eqx.Module):
    weight: jnp.ndarray
    bias: Optional[jnp.ndarray]
    lora_A: Optional[jnp.ndarray]
    lora_B: Optional[jnp.ndarray]
    alpha: float = 1.0
    rank: int = 0
    param_dtype: Any
    compute_dtype: Any

    def __init__(
        self,
        in_features,
        out_features,
        bias=False,
        rank=0,
        alpha=1.0,
        param_dtype=jnp.float32,
        compute_dtype=jnp.float32,
        key=None,
    ):
        self.param_dtype = param_dtype
        self.compute_dtype = compute_dtype
        if key is not None:
            keys = jax.random.split(key, 4)
        else:
            keys = jax.random.split(jax.random.PRNGKey(99), 4)

        self.weight = jax.random.normal(
            keys[0],
            (out_features, in_features),
            dtype=self.param_dtype,
        )
        self.bias = (
            jax.random.normal(keys[1], (out_features,), dtype=self.param_dtype)
            if bias
            else None
        )

        self.rank = rank
        self.alpha = alpha

        if rank > 0:
            # Initialize lora_A using Kaiming uniform initialization
            he_uniform = init.he_uniform()
            self.lora_A = he_uniform(
                keys[2], (in_features, rank), dtype=self.param_dtype
            )

            # Initialize lora_B with zeros
            self.lora_B = jnp.zeros((rank, out_features), dtype=self.param_dtype)
        else:
            self.lora_A = None
            self.lora_B = None

    def __call__(self, x):
        x = x.astype(self.compute_dtype)
        weight = self.weight.astype(self.compute_dtype)

        y = x @ weight.T
        if self.bias is not None:
            y += self.bias.astype(self.compute_dtype)
        if self.lora_A is not None and self.lora_B is not None:
            lora_A = self.lora_A.astype(self.compute_dtype)
            lora_B = self.lora_B.astype(self.compute_dtype)
            scaling = self.alpha / self.rank
            lora_update = (x @ lora_A @ lora_B) * scaling
            y += lora_update
        return y.astype(self.compute_dtype)


class LlamaRotaryEmbedding(eqx.Module):
    inv_freq: jnp.ndarray
    max_seq_len_cached: int
    param_dtype: Any
    compute_dtype: Any

    def __init__(
        self, config, param_dtype=jnp.float32, compute_dtype=jnp.float32
    ):
        self.param_dtype = param_dtype
        self.compute_dtype = compute_dtype

        dim = config.hidden_size // config.num_attention_heads
        self.max_seq_len_cached = config.max_position_embeddings
        inv_freq = 1.0 / (
            config.rope_theta
            ** (jnp.arange(0, dim, 2).astype(jnp.float32) / dim)
        )
        # TODO(mixed_precision): always using float32 for inv_freq.
        self.inv_freq = inv_freq.astype(jnp.float32)

    def __call__(self, x, position_ids):
        # TODO(mixed_precision): check if x should be retained as float32 for rotary embeddings.
        x = x.astype(jnp.float32)
        seq_len = position_ids.shape[1]
        t = position_ids.astype(jnp.float32)
        inv_freq = self.inv_freq

        # Reshape t to match the expected input shape
        t = t.reshape(-1, seq_len, 1)  # Shape: (batch_size, seq_len, 1)

        # Compute freqs directly without using einsum
        freqs = (
            t * inv_freq[None, None, :]
        )  # Shape: (batch_size, seq_len, dim//2)

        emb = jnp.concatenate(
            (freqs, freqs), axis=-1
        )  # Shape: (batch_size, seq_len, dim)
        cos = jnp.cos(emb)
        sin = jnp.sin(emb)
        return cos.astype(jnp.float32), sin.astype(jnp.float32)


class LlamaRMSNorm(eqx.Module):
    weight: jnp.ndarray
    eps: float
    param_dtype: Any
    compute_dtype: Any

    def __init__(
        self,
        hidden_size,
        eps=1e-6,
        param_dtype=jnp.float32,
        compute_dtype=jnp.float32,
    ):
        self.param_dtype = param_dtype
        self.compute_dtype = compute_dtype
        self.weight = jnp.ones(hidden_size, dtype=self.param_dtype)
        self.eps = eps

    def __call__(self, hidden_states):
        hidden_states = hidden_states.astype(self.compute_dtype)
        variance = jnp.mean(hidden_states**2, axis=-1, keepdims=True)
        hidden_states = hidden_states * jax.lax.rsqrt(variance + self.eps)
        return (self.weight.astype(self.compute_dtype) * hidden_states).astype(
            self.compute_dtype
        )


class LlamaSdpaAttention(eqx.Module):
    q_proj: LlamaLinear
    k_proj: LlamaLinear
    v_proj: LlamaLinear
    o_proj: LlamaLinear
    rotary_emb: LlamaRotaryEmbedding
    num_heads: int
    num_key_value_heads: int
    num_key_value_groups: int
    head_dim: int
    hidden_size: int
    max_position_embeddings: int
    rope_theta: float
    param_dtype: Any
    compute_dtype: Any

    def __init__(
        self,
        config,
        param_dtype=jnp.float32,
        compute_dtype=jnp.float32,
        key=None,
    ):
        self.param_dtype = param_dtype
        self.compute_dtype = compute_dtype
        if key is None:
            key = jax.random.PRNGKey(99)
        self.hidden_size = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.num_key_value_heads = config.num_key_value_heads
        self.num_key_value_groups = self.num_heads // self.num_key_value_heads
        self.head_dim = self.hidden_size // self.num_heads
        self.max_position_embeddings = config.max_position_embeddings
        self.rope_theta = config.rope_theta

        if (self.head_dim * self.num_heads) != self.hidden_size:
            raise ValueError(
                f"hidden_size must be divisible by num_heads (got `hidden_size`: {self.hidden_size}"
                f" and `num_heads`: {self.num_heads})."
            )

        keys = jax.random.split(key, 4)
        self.q_proj = LlamaLinear(
            self.hidden_size,
            self.num_heads * self.head_dim,
            bias=config.attention_bias,
            rank=config.lora_rank,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[0],
        )
        self.k_proj = LlamaLinear(
            self.hidden_size,
            self.num_key_value_heads * self.head_dim,
            bias=config.attention_bias,
            rank=config.lora_rank,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[1],
        )
        self.v_proj = LlamaLinear(
            self.hidden_size,
            self.num_key_value_heads * self.head_dim,
            bias=config.attention_bias,
            rank=config.lora_rank,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[2],
        )
        self.o_proj = LlamaLinear(
            self.num_heads * self.head_dim,
            self.hidden_size,
            bias=config.attention_bias,
            rank=config.lora_rank,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[3],
        )

        self.rotary_emb = LlamaRotaryEmbedding(
            config,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
        )

    def __call__(self, x, position_ids, attention_mask=None):
        x = x.astype(self.compute_dtype)

        def jax_apply_rotary_pos_emb(q, k, cos, sin):
            q_embed = (q * cos) + (jax_rotate_half(q) * sin)
            k_embed = (k * cos) + (jax_rotate_half(k) * sin)
            return q_embed, k_embed

        def jax_rotate_half(x):
            x1, x2 = jnp.split(x, 2, axis=-1)
            return jnp.concatenate((-x2, x1), axis=-1)

        bsz, q_len, _ = x.shape

        query_states = self.q_proj(x)
        key_states = self.k_proj(x)
        value_states = self.v_proj(x)

        query_states = query_states.reshape(
            bsz, q_len, self.num_heads, self.head_dim
        ).transpose(0, 2, 1, 3)
        key_states = key_states.reshape(
            bsz, q_len, self.num_key_value_heads, self.head_dim
        ).transpose(0, 2, 1, 3)
        value_states = value_states.reshape(
            bsz, q_len, self.num_key_value_heads, self.head_dim
        ).transpose(0, 2, 1, 3)

        cos, sin = self.rotary_emb(value_states, position_ids)

        # Reshape cos and sin to include num_heads dimension
        cos = cos[:, None, :, :]  # (batch_size, 1, seq_len, head_dim)
        sin = sin[:, None, :, :]  # (batch_size, 1, seq_len, head_dim)

        # Apply rotary embeddings
        query_states, key_states = jax_apply_rotary_pos_emb(
            query_states, key_states, cos, sin
        )

        if self.num_key_value_heads != self.num_heads:
            key_states = jnp.repeat(
                key_states, self.num_heads // self.num_key_value_heads, axis=1
            )
            value_states = jnp.repeat(
                value_states, self.num_heads // self.num_key_value_heads, axis=1
            )

        attn_weights = jnp.einsum(
            "bhqd,bhkd->bhqk", query_states, key_states
        ) / jnp.sqrt(self.head_dim)

        # Create causal mask
        q_len, k_len = query_states.shape[2], key_states.shape[2]
        causal_mask = jnp.tril(jnp.ones((q_len, q_len)))
        causal_mask = causal_mask[None, None, :, :]

        # If attention_mask is provided, use it to override the causal mask
        if attention_mask is not None:
            # Broadcast attention_mask to the correct shape
            attention_mask = jnp.expand_dims(attention_mask, axis=(1, 2))
            # Combine causal_mask and attention_mask
            combined_mask = jnp.minimum(causal_mask, attention_mask)
        else:
            combined_mask = causal_mask

        # Apply the combined mask
        attn_weights = jnp.where(combined_mask == 0, float("-inf"), attn_weights)

        attn_weights = jax.nn.softmax(attn_weights, axis=-1)

        attn_output = jnp.einsum("bhqk,bhkd->bhqd", attn_weights, value_states)
        attn_output = attn_output.transpose(0, 2, 1, 3).reshape(
            bsz, q_len, self.num_heads * self.head_dim
        )

        attn_output = self.o_proj(attn_output)

        return attn_output.astype(self.compute_dtype)


class LlamaMLP(eqx.Module):
    gate_proj: LlamaLinear
    up_proj: LlamaLinear
    down_proj: LlamaLinear
    param_dtype: Any
    compute_dtype: Any

    def __init__(
        self,
        hidden_size,
        intermediate_size,
        param_dtype=jnp.float32,
        compute_dtype=jnp.float32,
        key=None,
        rank=0,
    ):
        self.param_dtype = param_dtype
        self.compute_dtype = compute_dtype
        if key is None:
            key = jax.random.PRNGKey(99)
        keys = jax.random.split(key, 3)
        self.gate_proj = LlamaLinear(
            hidden_size,
            intermediate_size,
            bias=False,
            rank=rank,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[0],
        )
        self.up_proj = LlamaLinear(
            hidden_size,
            intermediate_size,
            bias=False,
            rank=rank,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[1],
        )
        self.down_proj = LlamaLinear(
            intermediate_size,
            hidden_size,
            bias=False,
            rank=rank,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[2],
        )

    def __call__(self, x):
        x = x.astype(self.compute_dtype)
        hidden = jax.nn.silu(self.gate_proj(x)) * self.up_proj(x)
        output = self.down_proj(hidden)
        return output.astype(self.compute_dtype)


class LlamaDecoderLayer(eqx.Module):
    self_attn: LlamaSdpaAttention
    mlp: LlamaMLP
    input_layernorm: LlamaRMSNorm
    post_attention_layernorm: LlamaRMSNorm
    param_dtype: Any
    compute_dtype: Any

    def __init__(
        self,
        config,
        param_dtype=jnp.float32,
        compute_dtype=jnp.float32,
        key=None,
    ):
        self.param_dtype = param_dtype
        self.compute_dtype = compute_dtype
        if key is None:
            key = jax.random.PRNGKey(99)
        keys = jax.random.split(key, 2)
        self.self_attn = LlamaSdpaAttention(
            config,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[0],
        )
        self.mlp = LlamaMLP(
            config.hidden_size,
            config.intermediate_size,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
            key=keys[1],
            rank=config.lora_rank,
        )
        self.input_layernorm = LlamaRMSNorm(
            config.hidden_size,
            config.rms_norm_eps,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
        )
        self.post_attention_layernorm = LlamaRMSNorm(
            config.hidden_size,
            config.rms_norm_eps,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
        )

    def __call__(self, hidden_states, attention_mask=None, position_ids=None):
        residual = hidden_states
        hidden_states = self.input_layernorm(hidden_states)
        hidden_states = self.self_attn(
            hidden_states, position_ids, attention_mask
        )
        hidden_states = residual + hidden_states

        residual = hidden_states
        hidden_states = self.post_attention_layernorm(hidden_states)
        hidden_states = self.mlp(hidden_states)
        hidden_states = residual + hidden_states

        return hidden_states.astype(self.compute_dtype)


class LlamaModel(eqx.Module):
    embed_tokens: LlamaEmbedding
    layers: List[LlamaDecoderLayer]
    norm: LlamaRMSNorm
    param_dtype: Any
    compute_dtype: Any

    def __init__(
        self,
        config,
        param_dtype=jnp.float32,
        compute_dtype=jnp.float32,
        key=None,
    ):
        self.param_dtype = param_dtype
        self.compute_dtype = compute_dtype
        if key is None:
            key = jax.random.PRNGKey(99)

        self.embed_tokens = LlamaEmbedding(
            config.vocab_size,
            config.hidden_size,
            param_dtype=jnp.float32,
            compute_dtype=jnp.float32,
            key=key,
        )

        layer_keys = jax.random.split(key, config.num_hidden_layers)
        # TODO(mfu): use lax.scan.
        self.layers = [
            LlamaDecoderLayer(
                config,
                param_dtype=self.param_dtype,
                compute_dtype=self.compute_dtype,
                key=layer_keys[i],
            )
            for i in range(config.num_hidden_layers)
        ]
        self.norm = LlamaRMSNorm(
            config.hidden_size,
            config.rms_norm_eps,
            param_dtype=self.param_dtype,
            compute_dtype=self.compute_dtype,
        )

    def __call__(self, input_ids, attention_mask=None, position_ids=None):
        hidden_states = self.embed_tokens(input_ids)

        policy = remat_policy["nothing"]
        for layer in self.layers:
            hidden_states = eqx.filter_checkpoint(layer, policy=policy)(
                hidden_states, attention_mask, position_ids
            )

        hidden_states = self.norm(hidden_states)

        return hidden_states.astype(self.compute_dtype)


class LlamaForCausalLM(eqx.Module):
    model: LlamaModel
    lm_head: LlamaLinear

    def __init__(
        self,
        config,
        param_dtype=jnp.float32,
        compute_dtype=jnp.float32,
        key=None,
    ):
        if key is None:
            key = jax.random.PRNGKey(99)
        key1, key2 = jax.random.split(key)
        self.model = LlamaModel(
            config,
            param_dtype=param_dtype,
            compute_dtype=compute_dtype,
            key=key1,
        )

        self.lm_head = LlamaLinear(
            config.hidden_size,
            config.vocab_size,
            bias=False,
            rank=config.lora_rank,
            param_dtype=param_dtype,
            compute_dtype=compute_dtype,
            key=key2,
        )

    def __call__(self, input_ids, attention_mask=None, position_ids=None):
        hidden_states = self.model(input_ids, attention_mask, position_ids)
        logits = self.lm_head(hidden_states)
        return logits.astype(self.model.compute_dtype)



from dataclasses import dataclass
from typing import Optional, Any, Tuple
import functools
import pyrallis
import jax

jax.distributed.initialize()

import equinox as eqx
import quax
import jax.numpy as jnp
from jax.experimental import mesh_utils
from jax.experimental.multihost_utils import (
    global_array_to_host_local_array,
    host_local_array_to_global_array,
)
import jax.tree_util as jtu
from jax.sharding import NamedSharding, PartitionSpec as PS
from src.felafax.trainer_engine.utils import named_tree_map

import optax
import os

from src.felafax.trainer_engine.checkpoint import (
    Checkpointer,
    load_llama_from_hf,
    save_model_to_hf,
    load_checkpoint_or_model,
)
from src.felafax.trainer_engine.models.llama3.jax.model import (
    LlamaForCausalLM,
    LlamaConfig,
    LlamaLinear,
)


def get_mesh(num_tpus: int, mesh_shape: Optional[Tuple[int, int, int]] = None):
    if mesh_shape is None:
        if num_tpus == 1:
            mesh_shape = (1, 1, 1)
        elif num_tpus == 2:
            mesh_shape = (1, 2, 1)
        elif num_tpus == 4:
            mesh_shape = (1, 2, 2)
        elif num_tpus == 8:
            mesh_shape = (2, 2, 2)
        elif num_tpus == 16:
            mesh_shape = (1, 4, 4)
        else:
            raise ValueError(f"Invalid number of TPUs: {num_tpus}")

    print(f"Creating TPU device mesh with shape {mesh_shape}...")
    device_mesh = mesh_utils.create_device_mesh(mesh_shape)
    mesh = jax.sharding.Mesh(device_mesh, axis_names=("batch", "fsdp", "mp"))
    return mesh


def merge_lora_params(model):
    def merge_fn(module):
        if (
            isinstance(module, LlamaLinear)
            and module.lora_A is not None
            and module.lora_B is not None
        ):
            delta_weight = module.lora_A @ module.lora_B.T
            new_weight = module.weight + delta_weight
            module = eqx.tree_at(lambda m: m.weight, module, new_weight)

            # Optionally set lora_A and lora_B to None
            module = eqx.tree_at(
                lambda m: (m.lora_A, m.lora_B), module, (None, None)
            )
        return module

    model = jtu.tree_map(merge_fn, model, is_leaf=eqx.is_array)
    return model


# Define configuration flags and default values
@dataclass
class TrainerConfig:
    """Configuration for the Llama trainer"""

    # Model configuration
    model_name: str = "meta-llama/Llama-3.2-1B"
    param_dtype: str = "float32"
    compute_dtype: str = "float32"

    # Training configuration
    num_epochs: int = 1
    num_steps: Optional[int] = None
    num_tpus: int = jax.device_count()
    mesh_shape: Optional[Tuple[int, int, int]] = None

    learning_rate: float = 1e-3

    # lora configuration
    lora_rank: int = 4  # Rank for lora matrices
    use_lora: bool = False  # Enable or disable lora training

    # Environment configuration
    base_dir: str = "/mnt/persistent-disk"
    hf_token: Optional[str] = None

    # Logging configuration
    log_interval: int = 10
    eval_interval: int = 10
    eval_steps: int = 10

    # Restore checkpoint
    restore_checkpoint: bool = False


# Core trainer class -- add non-essential things in private functions.
class Trainer:
    def __init__(
        self,
        trainer_config: TrainerConfig,
        train_dataloader: Any,
        val_dataloader: Any,
        model: Optional[eqx.Module] = None,
        model_config: Optional[LlamaConfig] = None,
        checkpointer: Optional[Checkpointer] = None,
    ):
        assert (
            trainer_config.model_name or model is not None
        ), "Either model_name must be provided in trainer_config or an existing model must be passed."

        self.trainer_config = trainer_config
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.mesh = get_mesh(trainer_config.num_tpus, trainer_config.mesh_shape)
        self.checkpointer = checkpointer

        if model is not None and model_config is not None:
            # Use the provided model and model_config
            self.model = model
            self.model_config = model_config
        elif checkpointer is not None and trainer_config.restore_checkpoint:
            print("Trying to restore checkpoint...")
            # Load from checkpoint if checkpointer is provided
            self.model, self.model_config = load_checkpoint_or_model(
                model_name=trainer_config.model_name,
                mesh=self.mesh,
                checkpointer=checkpointer,
                param_dtype=jnp.dtype(trainer_config.param_dtype),
                compute_dtype=jnp.dtype(trainer_config.compute_dtype),
            )
        else:
            print("Loading model from HuggingFace...")
            # Load the model and model_config from HuggingFace

            self.model, self.model_config = load_llama_from_hf(
                model_name=trainer_config.model_name,
                mesh=self.mesh,
                token=trainer_config.hf_token,
                lora_rank=self.trainer_config.lora_rank
                if self.trainer_config.use_lora
                else 0,
                param_dtype=jnp.dtype(trainer_config.param_dtype),
                compute_dtype=jnp.dtype(trainer_config.compute_dtype),
            )

        model_params, model_static = eqx.partition(self.model, eqx.is_array)

        if trainer_config.use_lora:
            # If using lora, create optimizer state only for the lora parameters.
            # Step 1: Create filter spec to identify lora params in the model pytree.
            self.is_lora_param_filter_spec = _make_lora_params_filter_spec(
                self.model
            )
            # Step 2: Partition the model parameters into lora and non-lora params.
            lora_params, _ = eqx.partition(
                model_params,
                filter_spec=self.is_lora_param_filter_spec,
                is_leaf=eqx.is_array,
            )
            optimizer_params = lora_params
            # TODO(lora): Add jax.lax.stop_gradient to non-lora params in the model tree.
        else:
            optimizer_params = model_params

        self.configure_optimizers(optimizer_params)

    def configure_optimizers(self, optimizer_params):
        self.optimizer = optax.chain(
            optax.clip_by_global_norm(1.0),  # Add gradient clipping
            optax.adam(learning_rate=self.trainer_config.learning_rate),
        )
        self.opt_state = self.optimizer.init(optimizer_params)

    @functools.partial(jax.jit, static_argnames=("self", "model_static"))
    def forward(self, model_params, model_static, batch):
        model = eqx.combine(model_params, model_static)
        input_ids = batch["input_ids"]
        labels = batch["labels"]
        attention_mask = batch.get("attention_mask", None)
        position_ids = batch.get("position_ids", None)

        logits = model(input_ids, attention_mask, position_ids)

        # Shift for next-token prediction
        shifted_logits = logits[..., :-1, :]
        shifted_labels = labels[..., 1:]

        # If using attention mask, shift it too.
        shifted_mask = None
        if attention_mask is not None:
            shifted_mask = attention_mask[..., 1:]

        loss, accuracy = _cross_entropy_loss_and_accuracy(
            shifted_logits, shifted_labels, shifted_mask
        )
        return loss, accuracy

    @functools.partial(
        jax.jit, static_argnames=("self", "model_static", "optimizer")
    )
    def training_step(
        self, model_params, model_static, optimizer, optimizer_state, batch
    ):
        grad_fn = jax.value_and_grad(self.forward, argnums=0, has_aux=True)
        (loss, accuracy), grads = grad_fn(
            model_params, model_static=model_static, batch=batch
        )

        if self.trainer_config.use_lora:
            # If using lora, only extract gradients for the lora params.
            # Step 1: Partition the gradients into lora and non-lora grads.
            lora_grads, _ = eqx.partition(
                grads,
                filter_spec=self.is_lora_param_filter_spec,
                is_leaf=eqx.is_array,
            )
            # Step 2: Partition the model parameters into lora and non-lora params.
            lora_params, non_lora_params = eqx.partition(
                model_params,
                filter_spec=self.is_lora_param_filter_spec,
                is_leaf=eqx.is_array,
            )
            # Step 3: Calculate updates to apply for each lora param using lora grad and apply them.
            updates, optimizer_state = optimizer.update(
                lora_grads, optimizer_state, lora_params
            )
            lora_params = optax.apply_updates(lora_params, updates)
            # Step 4: Combine the updated lora params and non-lora params.
            model_params = eqx.combine(lora_params, non_lora_params)
        else:
            # If not using lora, calculate updates to apply for all model params and apply them.
            updates, optimizer_state = optimizer.update(
                grads, optimizer_state, model_params
            )
            model_params = optax.apply_updates(model_params, updates)

        return loss, (accuracy, model_params, optimizer_state)

    @functools.partial(
        jax.jit,
        static_argnames=("self", "model_static"),
        donate_argnames=("batch",),
    )
    def validation_step(self, model_params, model_static, batch):
        model = eqx.combine(model_params, model_static)
        model = eqx.nn.inference_mode(model)

        input_ids = batch["input_ids"]
        labels = batch["labels"]
        attention_mask = batch.get("attention_mask", None)
        position_ids = batch.get("position_ids", None)

        logits = model(input_ids, attention_mask, position_ids)

        # Shift for next-token prediction
        shifted_logits = logits[..., :-1, :]
        shifted_labels = labels[..., 1:]

        # If using attention mask, shift it too.
        shifted_mask = None
        if attention_mask is not None:
            shifted_mask = attention_mask[..., 1:]

        loss, accuracy = _cross_entropy_loss_and_accuracy(
            shifted_logits, shifted_labels, shifted_mask
        )
        return loss, accuracy

    def train(self):
        model_params, model_static = eqx.partition(self.model, eqx.is_array)
        optimizer_state = self.opt_state
        max_steps = self.trainer_config.num_steps or float("inf")

        prev_step = 0
        prev_loss = 0.0
        prev_accuracy = 0.0
        prev_val_loss = 0.0
        prev_val_accuracy = 0.0

        for epoch in range(self.trainer_config.num_epochs):
            print(
                f"Started epoch {epoch + 1} of {self.trainer_config.num_epochs}..."
            )

            for step, batch in enumerate(self.train_dataloader):
                if step >= max_steps:
                    break

                if (
                    step == 1
                    or (step + 1) % self.trainer_config.log_interval == 0
                ):
                    # Printing metrics of previous step to avoid disrupting XLA pipelining
                    print(
                        f"Step {prev_step} | "
                        f"Train Loss: {prev_loss:.4f} | "
                        f"Val Loss: {prev_val_loss:.4f} | "
                        f"Next Token Prediction Accuracy (train, val): {prev_accuracy:.2%}, {prev_val_accuracy:.2%}"
                    )

                pass

                batch = _preprocess_batch(batch)
                batch = host_local_array_to_global_array(
                    batch, self.mesh, PS("batch")
                )
                optimizer_state = jax.device_put(
                    optimizer_state, NamedSharding(self.mesh, PS())
                )

                (
                    loss,
                    (accuracy, model_params, optimizer_state),
                ) = self.training_step(
                    model_params=model_params,
                    model_static=model_static,
                    optimizer=self.optimizer,
                    optimizer_state=optimizer_state,
                    batch=batch,
                )

                prev_step = step + 1
                prev_loss = loss
                prev_accuracy = accuracy

                if (
                    step == 0
                    or (step + 1) % self.trainer_config.eval_interval == 0
                ):
                    prev_val_loss, prev_val_accuracy = self.evaluate(
                        model_params=model_params,
                        model_static=model_static,
                        max_eval_steps=self.trainer_config.eval_steps,
                    )

                if self.checkpointer and (
                    (step + 1) % self.checkpointer.config.save_interval_steps
                    == 0
                ):
                    self.checkpointer.save_checkpoint(
                        model=eqx.combine(model_params, model_static),
                        model_config=self.model_config,
                        step=step + 1,
                    )

        # Update the model with the trained parameters
        self.model = eqx.combine(model_params, model_static)
        print("Training completed!")

        # Save final checkpoint
        if self.checkpointer:
            self.checkpointer.save_checkpoint(
                model=self.model,
                model_config=self.model_config,
                step=step + 1,
            )
            self.checkpointer.wait_until_finished()
            print("Final checkpoint saved at:", self.checkpointer.directory)

    def evaluate(self, model_params, model_static, max_eval_steps=None):
        """Run evaluation on the validation dataset.

        Args:
            model_params: The model parameters to evaluate
            model_static: The static model components
            max_eval_steps: Maximum number of evaluation steps (None for full dataset)

        Returns:
            tuple: (average_loss, average_accuracy)
        """
        max_eval_steps = max_eval_steps or float("inf")
        print(
            f"Running eval for {max_eval_steps if max_eval_steps != float('inf') else 'all'} steps..."
        )

        val_losses = []
        val_accuracies = []

        for eval_step, val_batch in enumerate(self.val_dataloader):
            if eval_step >= max_eval_steps:
                break

            val_batch = _preprocess_batch(val_batch)
            val_batch = jax.device_put(
                val_batch, NamedSharding(self.mesh, PS("batch"))
            )
            val_loss, val_accuracy = self.validation_step(
                model_params=model_params,
                model_static=model_static,
                batch=val_batch,
            )
            val_losses.append(val_loss)
            val_accuracies.append(val_accuracy)

        return (
            jnp.mean(jnp.array(val_losses)),
            jnp.mean(jnp.array(val_accuracies)),
        )

    def export(self, export_dir: Optional[str] = None):
        # After training, convert and save the model in Hugging Face format
        if self.trainer_config.use_lora:
            self.model = merge_lora_params(self.model)
        if export_dir is None:
            export_dir = os.path.join(self.trainer_config.base_dir, "hf_export")
        os.makedirs(export_dir, exist_ok=True)

        save_model_to_hf(
            model=self.model,
            model_config=self.model_config,
            output_dir=export_dir,
            tokenizer_name=self.trainer_config.model_name,
        )
        print("Hugging Face model saved at:", export_dir)


def _preprocess_batch(batch):
    # Convert PyTorch tensors to JAX arrays
    batch = {
        k: v if isinstance(v, jax.Array) else jax.numpy.array(v.numpy())
        for k, v in batch.items()
    }
    batch["input_ids"] = batch["input_ids"].astype(jnp.int32)
    batch["labels"] = batch["labels"].astype(jnp.int32)

    # Add position IDs to batch
    seq_length = batch["input_ids"].shape[1]
    batch["position_ids"] = jnp.repeat(
        jnp.arange(seq_length)[None, :],
        batch["input_ids"].shape[0],
        axis=0,
    )
    return batch


def _make_lora_params_filter_spec(model):
    return named_tree_map(
        lambda path_str, value: "lora_A" in path_str or "lora_B" in path_str,
        model,
        is_leaf=eqx.is_array,
    )


def _cross_entropy_loss_and_accuracy(logits, tokens, mask=None):
    if mask is None:
        mask = jnp.ones(tokens.shape[:2])
    mask = mask.astype(jnp.float32)

    valid_text_length = jnp.maximum(jnp.sum(mask, axis=-1), 1e-10)

    logits = logits.astype(jnp.float32)  # for numerical stability
    token_log_prob = jnp.squeeze(
        jnp.take_along_axis(
            jax.nn.log_softmax(logits, axis=-1),
            jnp.expand_dims(tokens, -1),
            axis=-1,
        ),
        -1,
    )
    token_log_prob = jnp.where(mask > 0.0, token_log_prob, jnp.array(0.0))
    loss = -jnp.mean(jnp.sum(token_log_prob, axis=-1) / valid_text_length)
    correct = jnp.where(
        mask > 0.0, jnp.argmax(logits, axis=-1) == tokens, jnp.array(False)
    )
    accuracy = jnp.mean(jnp.sum(correct, axis=-1) / valid_text_length)
    return loss, accuracy



"""Loads and saves checkpoint."""

import os
import json
import jax
import jax.numpy as jnp
import numpy as np
import torch

import equinox as eqx
import orbax.checkpoint as ocp
from transformers import LlamaForCausalLM as HFLlamaForCausalLM, AutoTokenizer
from src.felafax.trainer_engine.models.llama3.jax.model import (
    LlamaConfig,
    LlamaForCausalLM,
    LlamaLinear,
)

from typing import Optional, Tuple, Any
from jaxtyping import PyTree
from jax.sharding import NamedSharding, PartitionSpec as PS
from jax.experimental import mesh_utils

from dataclasses import dataclass
from .utils import named_tree_map


@dataclass
class CheckpointerConfig:
    """Configuration for checkpointing"""

    checkpoint_dir: str
    max_to_keep: int = 2
    save_interval_steps: int = 10
    create: bool = True  # Create the checkpoint directory if it doesn't exist
    enable_async_checkpointing: bool = True 


class Checkpointer:
    def __init__(self, config: CheckpointerConfig):
        if not config.checkpoint_dir:
            raise ValueError("Checkpoint directory cannot be empty")
        self.config = config
        self.checkpoint_dir = config.checkpoint_dir
        self.options = ocp.CheckpointManagerOptions(
            max_to_keep=config.max_to_keep,
            save_interval_steps=config.save_interval_steps,
            create=config.create,
            enable_async_checkpointing=config.enable_async_checkpointing,
        )

        self.checkpoint_mgr = ocp.CheckpointManager(
            directory=self.checkpoint_dir,
            options=self.options,
            item_names=["model_pytree", "model_config"],
        )

    @classmethod
    def get_abstract_pytree(cls, tree):
        return jax.tree_util.tree_map(ocp.utils.to_shape_dtype_struct, tree)

    def save_checkpoint(
        self, model: eqx.Module, model_config: LlamaConfig, step: int = 0
    ):
        """Saves model checkpoint."""
        model_pytree, _ = eqx.partition(model, eqx.is_inexact_array)
        self.checkpoint_mgr.save(
            step=step,
            args=ocp.args.Composite(
                model_pytree=ocp.args.StandardSave(model_pytree),
                model_config=ocp.args.JsonSave(model_config.to_dict()),
            ),
            force=True,
        )

    def restore_checkpoint(self) -> Tuple[eqx.Module, LlamaConfig]:
        """Restores model checkpoint."""
        # Step 1: Restore the model_config first
        restored_config = self.checkpoint_mgr.restore(
            step=self.checkpoint_mgr.latest_step(),
            items=["model_config"],
            args=ocp.args.Composite(
                model_config=ocp.args.JsonRestore(),
            ),
        )
        model_config = LlamaConfig(**restored_config["model_config"])

        # Step 2: Construct the model and create the abstract pytree
        model = LlamaForCausalLM(model_config)
        model_params, model_static = eqx.partition(model, eqx.is_inexact_array)
        model_abstract_pytree = self.get_abstract_pytree(model_params)

        # Step 3: Restore the model parameters using the abstract pytree
        restored_params = self.checkpoint_mgr.restore(
            step=self.checkpoint_mgr.latest_step(),
            items=["model_pytree"],
            args=ocp.args.Composite(
                model_pytree=ocp.args.StandardRestore(model_abstract_pytree),
            ),
        )

        # Combine restored model parameters with model static
        model_params = restored_params["model_pytree"]
        model = eqx.combine(model_params, model_static)
        return model, model_config

    def wait_until_finished(self):
        """Wait for any async operations to complete."""
        self.checkpoint_mgr.wait_until_finished()

    @property
    def directory(self):
        return self.checkpoint_mgr.directory


def load_model(
    model_name: str, mesh: jax.sharding.Mesh, token: Optional[str] = None
):
    """Loads a model from a checkpoint or Hugging Face.

    Args:
        model_name: Name or path of the model to load
        token: HuggingFace token for accessing gated models
    """
    return load_llama_from_hf(model_name, mesh=mesh, token=token)


def load_checkpoint_or_model(
    model_name: str,
    mesh: jax.sharding.Mesh,
    checkpointer: Checkpointer,
    param_dtype=jnp.float32,
    compute_dtype=jnp.float32,
) -> LlamaForCausalLM:
    """Loads checkpoint from local storage using Orbax or downloads from HF with specified dtypes.

    Args:
        model_name: Name of HF model (e.g. 'meta-llama/Llama-2-7b') or path to local checkpoint
        checkpointer: An instance of Checkpointer to manage loading
        param_dtype: The dtype in which parameters are stored and loaded
        output_dtype: The dtype in which computations are performed and outputs are returned

    Returns:
        tuple: (model, model_config)
    """
    has_checkpoints = len(checkpointer.checkpoint_mgr.all_steps()) > 0
    if has_checkpoints:
        # Restores the model in whatever dtypes are stored in the checkpoint.
        model, model_config = checkpointer.restore_checkpoint()
        print(
            f"Restored checkpoint from step {checkpointer.checkpoint_mgr.latest_step()}"
        )
        return model, model_config

    model, model_config = load_llama_from_hf(
        model_name,
        mesh=mesh,
        param_dtype=param_dtype,
        compute_dtype=compute_dtype,
    )
    return model, model_config


def create_llama_config_from_hf_model(hf_model) -> LlamaConfig:
    """Creates Equinox config from Hugging Face model config."""
    return LlamaConfig(
        vocab_size=hf_model.config.vocab_size,
        hidden_size=hf_model.config.hidden_size,
        intermediate_size=hf_model.config.intermediate_size,
        num_hidden_layers=hf_model.config.num_hidden_layers,
        num_attention_heads=hf_model.config.num_attention_heads,
        num_key_value_heads=hf_model.config.num_key_value_heads,
        max_position_embeddings=hf_model.config.max_position_embeddings,
        rms_norm_eps=hf_model.config.rms_norm_eps,
        rope_theta=hf_model.config.rope_theta,
        attention_bias=hf_model.config.attention_bias,
    )


def _make_torch_to_jax(dtype, mesh):
    """Creates a closure that converts PyTorch tensors to JAX arrays with sharding annotations."""

    def _torch_to_jax(tensor, sharding_spec):
        jax_array = jnp.array(tensor.detach().numpy(), dtype=dtype)
        sharding = NamedSharding(mesh, sharding_spec)
        return jax.device_put(jax_array, sharding)

    return _torch_to_jax


# TODO(refactor): Move load model into models/llama.
def load_llama_from_hf(
    model_name: str,
    mesh: jax.sharding.Mesh,
    token: Optional[str] = None,
    lora_rank: int = 0,
    param_dtype: Any = jnp.float32,
    compute_dtype: Any = jnp.float32,
) -> LlamaForCausalLM:
    """Downloads and converts Hugging Face model to Equinox model with specified dtypes.

    Args:
        model_name: Name of the Hugging Face model to load
        token: HuggingFace token for accessing gated models
        lora_rank: Rank for LoRA parameters (set to 0 for no LoRA)
        param_dtype: The dtype in which parameters are stored
        output_dtype: The dtype in which computations are performed

    Returns:
        eqx_model: LlamaForCausalLM model with specified dtypes
        model_config: Configuration of the model
    """
    # Load HF model
    hf_model = HFLlamaForCausalLM.from_pretrained(
        model_name, torch_dtype=torch.float32, token=token
    )

    # Create config and initialize Equinox model
    model_config = create_llama_config_from_hf_model(hf_model)
    model_config.lora_rank = lora_rank

    key = jax.random.PRNGKey(99)
    eqx_model = LlamaForCausalLM(
        model_config,
        param_dtype=param_dtype,
        compute_dtype=compute_dtype,
        key=key,
    )
    torch_to_jax_float32 = _make_torch_to_jax(dtype=jnp.float32, mesh=mesh)
    torch_to_jax = _make_torch_to_jax(dtype=param_dtype, mesh=mesh)

    # Copy weights from HF model to Equinox model
    eqx_model = eqx.tree_at(
        lambda t: t.model.embed_tokens.weight,
        eqx_model,
        # Copy embedding weights at float32 precision.
        torch_to_jax_float32(
            hf_model.model.embed_tokens.weight, PS(("mp", "fsdp"))
        ),
    )
    eqx_model = eqx.tree_at(
        lambda t: t.model.norm.weight,
        eqx_model,
        torch_to_jax(hf_model.model.norm.weight, PS()),
    )
    eqx_model = eqx.tree_at(
        lambda t: t.lm_head.weight,
        eqx_model,
        torch_to_jax(hf_model.lm_head.weight, PS(("fsdp", "mp"))),
    )

    # Copy layer weights with appropriate sharding
    for i in range(len(eqx_model.model.layers)):
        hf_layer = hf_model.model.layers[i]

        # Self-attention weights
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].self_attn.q_proj.weight,
            eqx_model,
            torch_to_jax(hf_layer.self_attn.q_proj.weight, PS(("fsdp", "mp"))),
        )
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].self_attn.k_proj.weight,
            eqx_model,
            torch_to_jax(hf_layer.self_attn.k_proj.weight, PS(("fsdp", "mp"))),
        )
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].self_attn.v_proj.weight,
            eqx_model,
            torch_to_jax(hf_layer.self_attn.v_proj.weight, PS(("fsdp", "mp"))),
        )
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].self_attn.o_proj.weight,
            eqx_model,
            torch_to_jax(hf_layer.self_attn.o_proj.weight, PS(("mp", "fsdp"))),
        )

        # MLP weights
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].mlp.gate_proj.weight,
            eqx_model,
            torch_to_jax(hf_layer.mlp.gate_proj.weight, PS(("fsdp", "mp"))),
        )
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].mlp.up_proj.weight,
            eqx_model,
            torch_to_jax(hf_layer.mlp.up_proj.weight, PS(("fsdp", "mp"))),
        )
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].mlp.down_proj.weight,
            eqx_model,
            torch_to_jax(hf_layer.mlp.down_proj.weight, PS(("mp", "fsdp"))),
        )

        # Layer norms
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].input_layernorm.weight,
            eqx_model,
            torch_to_jax(hf_layer.input_layernorm.weight, PS()),
        )
        eqx_model = eqx.tree_at(
            lambda t: t.model.layers[i].post_attention_layernorm.weight,
            eqx_model,
            torch_to_jax(hf_layer.post_attention_layernorm.weight, PS()),
        )

    return eqx_model, model_config


def save_model_to_hf(
    model: eqx.Module,
    model_config: LlamaConfig,
    output_dir: str,
    tokenizer_name: str = "meta-llama/Llama-2-7b-hf",
):
    """Converts Equinox model back to Hugging Face format and saves it.

    Args:
        model: Equinox LlamaForCausalLM model
        model_config: LlamaConfig used for the model
        output_dir: Directory where to save the Hugging Face model
        tokenizer_name: Name of the tokenizer to download and save
    """
    import torch
    import numpy as np
    from transformers import LlamaForCausalLM as HFLlamaForCausalLM
    from transformers import LlamaConfig as HFLlamaConfig
    from transformers import AutoTokenizer

    # Ensure the output directory exists
    os.makedirs(output_dir, exist_ok=True)

    # Create a Hugging Face config from Equinox config
    hf_config = HFLlamaConfig(
        vocab_size=model_config.vocab_size,
        hidden_size=model_config.hidden_size,
        intermediate_size=model_config.intermediate_size,
        num_hidden_layers=model_config.num_hidden_layers,
        num_attention_heads=model_config.num_attention_heads,
        num_key_value_heads=model_config.num_key_value_heads,
        max_position_embeddings=model_config.max_position_embeddings,
        rms_norm_eps=model_config.rms_norm_eps,
        rope_theta=model_config.rope_theta,
        attention_bias=model_config.attention_bias,
    )

    # Initialize a Hugging Face model with the same configuration
    hf_model = HFLlamaForCausalLM(config=hf_config)

    # Remove sharding and convert JAX arrays to NumPy arrays
    model_params, _ = eqx.partition(model, eqx.is_array)
    model_params = jax.tree_util.tree_map(
        lambda x: np.array(x).astype(np.float32),
        model_params,
    )

    # Copy weights from Equinox model to Hugging Face model
    # Embedding weights
    hf_model.model.embed_tokens.weight.data = torch.tensor(
        model_params.model.embed_tokens.weight, dtype=torch.float32
    )
    hf_model.lm_head.weight.data = torch.tensor(
        model_params.lm_head.weight, dtype=torch.float32
    )
    hf_model.model.norm.weight.data = torch.tensor(
        model_params.model.norm.weight, dtype=torch.float32
    )

    # Layer-wise weights
    for i in range(len(hf_model.model.layers)):
        eqx_layer = model_params.model.layers[i]
        hf_layer = hf_model.model.layers[i]

        # Self-attention weights
        hf_layer.self_attn.q_proj.weight.data = torch.tensor(
            eqx_layer.self_attn.q_proj.weight, dtype=torch.float32
        )
        hf_layer.self_attn.k_proj.weight.data = torch.tensor(
            eqx_layer.self_attn.k_proj.weight, dtype=torch.float32
        )
        hf_layer.self_attn.v_proj.weight.data = torch.tensor(
            eqx_layer.self_attn.v_proj.weight, dtype=torch.float32
        )
        hf_layer.self_attn.o_proj.weight.data = torch.tensor(
            eqx_layer.self_attn.o_proj.weight, dtype=torch.float32
        )

        # MLP weights
        hf_layer.mlp.gate_proj.weight.data = torch.tensor(
            eqx_layer.mlp.gate_proj.weight, dtype=torch.float32
        )
        hf_layer.mlp.up_proj.weight.data = torch.tensor(
            eqx_layer.mlp.up_proj.weight, dtype=torch.float32
        )
        hf_layer.mlp.down_proj.weight.data = torch.tensor(
            eqx_layer.mlp.down_proj.weight, dtype=torch.float32
        )

        # Layer norms
        hf_layer.input_layernorm.weight.data = torch.tensor(
            eqx_layer.input_layernorm.weight, dtype=torch.float32
        )
        hf_layer.post_attention_layernorm.weight.data = torch.tensor(
            eqx_layer.post_attention_layernorm.weight, dtype=torch.float32
        )

    # Save the Hugging Face model
    hf_model.save_pretrained(output_dir)

    # Save the tokenizer
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)
    tokenizer.save_pretrained(output_dir)

    print(f"Model and tokenizer saved to {output_dir}")



from huggingface_hub import create_repo, HfApi, HfFolder, upload_folder
import os
import jax

from typing import Optional


def upload_dir_to_hf(
    dir_path: str,
    repo_name: str,
    commit_message: str = "Add fine-tuned model",
    token: Optional[str] = None,
):
    """
    Uploads the model and tokenizer to Hugging Face Hub.

    Args:
        output_dir: Directory containing the model and tokenizer files.
        repo_name: Name of the repository on Hugging Face Hub (e.g., 'username/repo_name').
        commit_message: Commit message for the upload.
        token: Optional HF token for authentication.
    """
    api = HfApi()
    token = HfFolder.get_token() if token is None else token
    if token is None:
        raise ValueError(
            "You must either provide a token or be logged in to HuggingFace Hub to upload models."
        )

    # Create repo if it doesn't exist
    api.create_repo(repo_id=repo_name, exist_ok=True, token=token)

    # Upload the folder
    upload_folder(
        repo_id=repo_name,
        folder_path=dir_path,
        commit_message=commit_message,
        token=token,
        ignore_patterns=["*.py"],
    )

    print(
        f"Model uploaded to Hugging Face Hub at https://huggingface.co/{repo_name}"
    )


def named_tree_map(f, tree, is_leaf=None, sep="/"):
    """Maps a function over a JAX tree, providing path strings and values to the function.

    This function traverses a JAX tree structure and applies a function to each node,
    passing both the string representation of the node's path and its value.

    Args:
        f: Function taking (path_str, value) arguments to apply at each node
        tree: JAX tree structure to traverse
        is_leaf: Optional function to determine leaf nodes
        sep: Separator for path components (e.g., 'a/b/c' if sep='/')

    Example:
        >>> def print_node(path, val): print(f"{path}: {val}")
        >>> tree = {"layer1": {"w": 1, "b": 2}}
        >>> named_tree_map(print_node, tree)
        layer1/w: 1
        layer1/b: 2
    """

    def convert_path_key(key):
        """Converts a single path key to its string representation."""
        if isinstance(key, jax.tree_util.SequenceKey):
            return str(key.idx)
        elif isinstance(key, jax.tree_util.DictKey):
            return str(key.key)
        elif isinstance(key, jax.tree_util.GetAttrKey):
            return str(key.name)
        elif isinstance(key, jax.tree_util.FlattenedIndexKey):
            return str(key.key)
        return str(key)

    def process_node(path, value):
        """Processes a single node by converting its path and conditionally applying f."""
        path_str = sep.join(convert_path_key(k) for k in path)
        if is_leaf(value):
            return f(path_str, value)
        else:
            # TODO(lora): fix this
            return f(path_str, value)
            # return value

    return jax.tree_util.tree_map_with_path(process_node, tree, is_leaf=is_leaf)



# base.py
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union, Tuple

import torch
from datasets import load_dataset
from torch.utils.data import DataLoader, Dataset
from transformers import PreTrainedTokenizerBase


@dataclass
class DatasetConfig:
    """Configuration for datasets."""

    # Data loading parameters
    data_source: str = ""
    max_examples: Optional[int] = None
    split: str = "train"
    train_test_split: float = 0.15
    seed: int = 42

    # Processing parameters
    batch_size: int = 32
    max_seq_length: int = 64
    num_workers: int = 4
    ignore_index: int = -100
    mask_prompt: bool = False
    pad_id: int = 0

    # Optional transform function for custom preprocessing
    transform: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None


def load_data(
    config: DatasetConfig,
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    """Loads the dataset and returns train and validation splits."""

    # Load dataset from Hugging Face Hub or local file
    if Path(config.data_source).is_file():
        dataset = load_dataset(
            "json",
            data_files=config.data_source,
            split=config.split,
        )
    else:
        dataset = load_dataset(
            config.data_source,
            split=config.split,
        )

    # If max_examples is set, limit the number of examples
    if config.max_examples is not None:
        dataset = dataset.select(range(min(config.max_examples, len(dataset))))

    # Split into train and validation sets
    dataset = dataset.train_test_split(
        test_size=config.train_test_split,
        seed=config.seed,
    )

    train_data = [sample for sample in dataset["train"]]
    val_data = [sample for sample in dataset["test"]]

    return train_data, val_data


def create_dataloader(
    config: DatasetConfig,
    dataset: "SFTDataset",
    shuffle: bool = False,
) -> DataLoader:
    """Creates a DataLoader for the given dataset."""
    return DataLoader(
        dataset,
        batch_size=config.batch_size,
        shuffle=shuffle,
        num_workers=config.num_workers,
        collate_fn=get_sft_collate_fn(
            max_seq_length=config.max_seq_length,
            pad_id=config.pad_id,
            ignore_index=config.ignore_index,
        ),
    )


class SFTDataset(Dataset):
    """Dataset for Supervised Fine-Tuning (SFT)."""

    def __init__(
        self,
        config: DatasetConfig,
        data: List[Dict[str, Any]],
        tokenizer: PreTrainedTokenizerBase,
    ):
        self.data = data
        self.tokenizer = tokenizer
        self.config = config

        self.max_seq_length = config.max_seq_length
        self.mask_prompt = config.mask_prompt
        self.ignore_index = config.ignore_index
        self.transform = config.transform
        self.eos_token_id = (
            self.tokenizer.eos_token_id
            if self.tokenizer.eos_token_id is not None
            else self.tokenizer.pad_token_id
        )

    def apply_format(self, example: Dict[str, Any]) -> Tuple[str, str]:
        """Default method to apply prompt formatting. Returns prompt and response.
        Override this method in subclasses for custom behavior."""
        prompt = (
            "Below is an instruction that describes a task. "
            "Write a response that appropriately completes the request.\n\n"
            f"### Instruction:\n{example['instruction']}\n\n### Response:\n"
        )
        response = example["output"]
        return prompt, response

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        example = self.data[idx]

        if self.transform:
            example = self.transform(example)

        # Apply the prompt formatting
        prompt, response = self.apply_format(example)

        # Encode the prompt with special tokens
        encoded_prompt = self.tokenizer.encode(
            prompt,
            add_special_tokens=True,
            max_length=self.max_seq_length,
            truncation=True,
        )

        # Encode the response without adding special tokens
        encoded_response = self.tokenizer.encode(
            response,
            add_special_tokens=False,
            max_length=self.max_seq_length,
            truncation=True,
        )

        # Concatenate the encoded prompt and response
        encoded_prompt_and_response = (
            encoded_prompt + encoded_response + [self.eos_token_id]
        )

        # Truncate the combined sequence to the max_seq_length if necessary
        if self.max_seq_length > 0:
            encoded_prompt_and_response = encoded_prompt_and_response[
                : self.max_seq_length
            ]

        # Convert to torch tensor
        input_ids = torch.tensor(encoded_prompt_and_response, dtype=torch.long)

        # Create labels, masking the prompt if required
        labels = input_ids.clone()
        if self.mask_prompt:
            labels[: len(encoded_prompt)] = self.ignore_index

        return {
            "input_ids": input_ids,
            "labels": labels,
            "prompt_length": len(encoded_prompt),
            "response_length": len(encoded_response),
        }


def get_sft_collate_fn(
    max_seq_length: int = -1,
    pad_id: int = 0,
    ignore_index: int = -100,
) -> Callable:
    """Returns the collate function for supervised fine-tuning."""
    return partial(
        _sft_collate_fn,
        max_seq_length=max_seq_length,
        pad_id=pad_id,
        ignore_index=ignore_index,
    )


def _sft_collate_fn(
    samples: List[Dict[str, Any]],
    max_seq_length: int,
    pad_id: int = 0,
    ignore_index: int = -100,
) -> Dict[str, torch.Tensor]:
    """Collate function that pads sequences to max_seq_length."""
    batched = {}
    for key in ("input_ids", "labels"):
        pad_value = pad_id if key == "input_ids" else ignore_index

        # Truncate and pad sequences
        sequences = [sample[key][:max_seq_length] for sample in samples]
        padded_sequences = [
            torch.nn.functional.pad(
                seq, (0, max_seq_length - len(seq)), value=pad_value
            )
            for seq in sequences
        ]
        batched[key] = torch.stack(padded_sequences)

    # Process lengths
    for key in ("prompt_length", "response_length"):
        lengths = torch.tensor(
            [min(sample[key], max_seq_length) for sample in samples],
            dtype=torch.int64,
        ).unsqueeze(1)

        batched[key] = lengths

    return batched



"""Sets up the environment for the training and provides utility functions."""

import os
import jax
import warnings


def setup_environment(trainer_config):
    """
    Sets up the environment variables and JAX configurations.

    Args:
        trainer_config (TrainerConfig): Configuration object containing environment settings.
    """
    base_dir = trainer_config.base_dir

    # Setting up Hugging Face directories
    hf_dir = os.path.join(base_dir, "hf")
    os.makedirs(hf_dir, exist_ok=True)
    os.environ["HF_HUB_CACHE"] = hf_dir
    os.environ["HF_HOME"] = hf_dir
    os.environ["TOKENIZERS_PARALLELISM"] = "false"

    # Setting up JAX cache directories and configurations
    jax_cache_dir = os.path.join(base_dir, "jax_cache")
    os.makedirs(jax_cache_dir, exist_ok=True)
    jax.config.update("jax_compilation_cache_dir", jax_cache_dir)
    jax.config.update("jax_persistent_cache_min_entry_size_bytes", -1)
    jax.config.update("jax_persistent_cache_min_compile_time_secs", 0)
    jax.config.update("jax_threefry_partitionable", True)

    # To suppress warning: RuntimeWarning: os.fork() was called. os.fork() is incompatible with multithreaded code, and JAX is multithreaded, so this will likely lead to a deadlock.
    warnings.filterwarnings("ignore", message="os.fork()")
